"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.canBeFalsyWhenNotNull = exports.canBeFalsy = exports.isFunctionType = exports.isArrayType = exports.forTypeOrAnySupertype = exports.isNullableType = exports.isNumberType = exports.isStringType = exports.isNullishType = exports.typeCanSatisfy = exports.typeAlwaysSatisfies = exports.isTypeWithFlags = void 0;
const ts = require("typescript");
function isTypeWithFlags(context, type, flags) {
    const predicate = (type) => (type.flags & flags) !== 0;
    return typeAlwaysSatisfies(context, type, predicate);
}
exports.isTypeWithFlags = isTypeWithFlags;
function typeAlwaysSatisfies(context, type, predicate) {
    const baseConstraint = context.checker.getBaseConstraintOfType(type);
    if (baseConstraint) {
        type = baseConstraint;
    }
    if (predicate(type)) {
        return true;
    }
    if (type.isUnion()) {
        return type.types.every(t => typeAlwaysSatisfies(context, t, predicate));
    }
    if (type.isIntersection()) {
        return type.types.some(t => typeAlwaysSatisfies(context, t, predicate));
    }
    return false;
}
exports.typeAlwaysSatisfies = typeAlwaysSatisfies;
function typeCanSatisfy(context, type, predicate) {
    const baseConstraint = context.checker.getBaseConstraintOfType(type);
    if (!baseConstraint) {
        // type parameter with no constraint can be anything, assume it might satisfy predicate
        if (type.isTypeParameter())
            return true;
    }
    else {
        type = baseConstraint;
    }
    if (predicate(type)) {
        return true;
    }
    if (type.isUnion()) {
        return type.types.some(t => typeCanSatisfy(context, t, predicate));
    }
    if (type.isIntersection()) {
        return type.types.some(t => typeCanSatisfy(context, t, predicate));
    }
    return false;
}
exports.typeCanSatisfy = typeCanSatisfy;
function isNullishType(context, type) {
    return isTypeWithFlags(context, type, ts.TypeFlags.Undefined | ts.TypeFlags.Null | ts.TypeFlags.VoidLike);
}
exports.isNullishType = isNullishType;
function isStringType(context, type) {
    return isTypeWithFlags(context, type, ts.TypeFlags.String | ts.TypeFlags.StringLike | ts.TypeFlags.StringLiteral);
}
exports.isStringType = isStringType;
function isNumberType(context, type) {
    return isTypeWithFlags(context, type, ts.TypeFlags.Number | ts.TypeFlags.NumberLike | ts.TypeFlags.NumberLiteral);
}
exports.isNumberType = isNumberType;
function isNullableType(context, type, isType) {
    return (typeCanSatisfy(context, type, t => isType(context, t)) &&
        typeAlwaysSatisfies(context, type, t => isType(context, t) || isNullishType(context, t)));
}
exports.isNullableType = isNullableType;
function isExplicitArrayType(context, type) {
    if (type.symbol) {
        const baseConstraint = context.checker.getBaseConstraintOfType(type);
        if (baseConstraint && baseConstraint !== type) {
            return isExplicitArrayType(context, baseConstraint);
        }
    }
    if (type.isUnionOrIntersection()) {
        return type.types.some(t => isExplicitArrayType(context, t));
    }
    const flags = ts.NodeBuilderFlags.InTypeAlias | ts.NodeBuilderFlags.AllowEmptyTuple;
    let typeNode = context.checker.typeToTypeNode(type, undefined, flags);
    if (typeNode && ts.isTypeOperatorNode(typeNode) && typeNode.operator === ts.SyntaxKind.ReadonlyKeyword) {
        typeNode = typeNode.type;
    }
    return typeNode !== undefined && (ts.isArrayTypeNode(typeNode) || ts.isTupleTypeNode(typeNode));
}
/**
 * Iterate over a type and its bases until the callback returns true.
 */
function forTypeOrAnySupertype(context, type, predicate) {
    var _a;
    if (predicate(type)) {
        return true;
    }
    if (!type.isClassOrInterface() && type.symbol) {
        type = context.checker.getDeclaredTypeOfSymbol(type.symbol);
    }
    return ((_a = type.getBaseTypes()) !== null && _a !== void 0 ? _a : []).some(superType => forTypeOrAnySupertype(context, superType, predicate));
}
exports.forTypeOrAnySupertype = forTypeOrAnySupertype;
function isArrayType(context, type) {
    return forTypeOrAnySupertype(context, type, t => isExplicitArrayType(context, t));
}
exports.isArrayType = isArrayType;
function isFunctionType(type) {
    return type.getCallSignatures().length > 0;
}
exports.isFunctionType = isFunctionType;
function canBeFalsy(context, type) {
    const strictNullChecks = context.options.strict === true || context.options.strictNullChecks === true;
    const falsyFlags = ts.TypeFlags.Boolean |
        ts.TypeFlags.BooleanLiteral |
        ts.TypeFlags.Never |
        ts.TypeFlags.Void |
        ts.TypeFlags.Unknown |
        ts.TypeFlags.Any |
        ts.TypeFlags.Undefined |
        ts.TypeFlags.Null;
    return typeCanSatisfy(context, type, type => (type.flags & falsyFlags) !== 0 || (!strictNullChecks && !type.isLiteral()));
}
exports.canBeFalsy = canBeFalsy;
function canBeFalsyWhenNotNull(context, type) {
    const falsyFlags = ts.TypeFlags.Boolean |
        ts.TypeFlags.BooleanLiteral |
        ts.TypeFlags.Never |
        ts.TypeFlags.Void |
        ts.TypeFlags.Unknown |
        ts.TypeFlags.Any;
    return typeCanSatisfy(context, type, type => (type.flags & falsyFlags) !== 0);
}
exports.canBeFalsyWhenNotNull = canBeFalsyWhenNotNull;
//# sourceMappingURL=types.js.map